#include "lip.h"                         /*  Date:  2007-12-16   Version  2.4 */
#include "upr.h"
#include <stdlib.h>   /*  void exit() */
#include <unistd.h>   /** only for getpid(), for random generator initialize **/
#ifndef  PSIEVE_SIZE
#define  PSIEVE_SIZE  (1<<24)   /*  30*PSIEVE_SIZE  must be < max unsigned  */
#endif
#define  BIG_KNOWN_ASCII_FACTORS  "known_factor"
static long unique = 0;
static long grow = 2;
static long nptests = 30;
static long ecmattempts = 1000;
static long ecmbound = 10000;
static long smallpbound = 100000;
static FILE *certificat= (FILE*)0;

static long ecm_primefactor(verylong n, verylong *f)
   /****************************************************\
   *  f will be set to a prime divisor of n
   *  returns 0 if successfull
   *  returns < 0 if error occurs
   *  returns 1  if  n <= 1
   \****************************************************/
{
    long  i, j, result;
    static verylong  a=0, c=0, r=0;

    unique =  /* 5412389 */ getpid();
    {   FILE  *hp = fopen(BIG_KNOWN_ASCII_FACTORS,"r");
        while (hp && 1==zfread(hp,&a) &&  1==zfread(hp,f) )
        {
            zdiv(a,*f,&c,&r);
            if (!ziszero(r))
               fprintf(stderr,"wrong factor in %s\n",BIG_KNOWN_ASCII_FACTORS),  exit(11);
            if (!zcompare(n,a)  &&  zprime(*f,nptests,2))
            {
               fprintf(stderr,"known factor of: "); zfwriteln(stderr,a);
               return 0;
            }
        }
        if (hp)
            fclose(hp);
    }
    if (zscompare(n,1)<=0)  return  1;
    zcopy(n,&a);
    for (;;)
    {
#ifdef  ECM_CHECK
        fprintf(stderr,"to factor: "); zfwrite(stderr,a);
#endif
        result = zecm(a,f,unique,ecmattempts,ecmbound,grow,0,0);
#ifdef  ECM_CHECK
        fprintf(stderr,"\n");
#endif
        if (result <0)
            return  -1;    /**  "bug, no factor!"  **/
        else if (result==2) 
        {   zsqrt(a,f,&r);
            if (zscompare(r,0)==0 && zprime(*f,nptests,2))
               return  0;
            for (i=3;zscompare(*f,2)<0;i++)
            {  zroot(a,i,f);
               zcopy(*f,&c);
               for (j=1;j<i;j++)
                  zmulin(*f,&c);
               if (zcompare(a,c)==0 && zprime(*f,nptests,2))
                  return  0;
            }
            return  -2;   /**   "prime power error"   **/
        }
        else if (result==1)
        {   if (zprime(*f,nptests,2))
               return  0;
            else
            {  zdiv(a,*f,&c,&r);
               if (zprime(c,nptests,2))
               {  zcopy(c,f);
                  return  0;
               }
            }
            zcopy(*f,&a);
            if (unique > 0) unique ++; else unique --;
        } 
        else if (result==0)
        {   if (zprime(a,nptests,2))
            {  zcopy(a,f);
               return  0;
            }
            if (unique > 0) unique ++; else unique --;
        }
    }
}

long  factor_control(long  param[])
   /****************************************************\
   *  changes the default internal ecm parameters
   *  returns  0   if  all parameters are in valid range
   *  a parameter value of 0 means: don't change this parameter
   \****************************************************/
{  if (param[0] < 3)  return 1;  /* bound to test for small prime factors by trial-division */
   if (param[1] < 0)  return 2;  /* bound of first curve in phase 1 */
   if (param[2] < 0)  return 3;  /* at most number of curves to try */
   if (param[3] < 0)  return 4;  /* growrate is 1+grow/100  */
   if (param[4] < 0)  return 5;  /* number of probable prime tests */
   if (param[0])
      smallpbound= param[0];
   if (param[1])
      ecmbound= param[1];
   if (param[2])
      ecmattempts= param[2];
   if (param[3])
      grow= param[3];
   if (param[4])
      nptests= param[4];       /**  >= ~20  **/
   return  0;
}

long factor(verylong n, verylong pri[], unsigned exp[])
   /****************************************************\
   *  factorizes n completely
   *  stores in pri the different primefactors and in exp their exponents
   *  returns  the number of different primefactors
   *  returns  0   if  n ==  -1, 0, 1
   *  returns  -1  if  internal factoring fails
   \****************************************************/
{
   long  i, j, k;
   static verylong  a=0, c=0, d=0, r=0, w=0;
   zcopy(n,&a);
   if (zscompare(a,0)<0)  znegate(&a);
   if (zscompare(a,1)<=0) return  0;
   k = 0;
   for (j=2;j<4;j++)
   {  if (zsdiv(a,j,&c)==0)
      {  zintoz(j,&pri[k]);
         exp[k] = 0;
         do {
            exp[k] ++;
            zcopy(c,&a);
         } while (zsdiv(a,j,&c)==0);
         k++;
      }
   }
   zsqrt(a,&w,&c);
   j=5;
   for (;;)
   {  i = zscompare(w,smallpbound) < 0 ? ztoint(w) : smallpbound;
      if (j > i)  break;
      if ( (j = ztridiv(a, &c, j, i)) > i)  break;
      zintoz(j,&pri[k]);
      exp[k] = 0;
      do {
         exp[k] ++;
         zcopy(c,&a);
      } while (zsdiv(a,j,&c)==0);
      k++;
      zsqrt(a,&w,&c);
      j += 2;
   }
   if (zscompare(w,j) >= 0)
   {  while ( ! zprime(a,nptests,2) ) 
      {
         if (ecm_primefactor(a,&d))
            return  fprintf(stderr,"zecm in ecm_primefactor failed!\n"),  -1;
         for (i=k;i--;)
            if (zcompare(d,pri[i]) > 0)  break;
         i++;
         for (j=k;j>i;j--)
         {  zcopy(pri[j-1],&pri[j]);
            exp[j]=exp[j-1];
         }
         zcopy(d,&pri[i]);
         exp[i] = 0;
         do {
            exp[i] ++;
            zdiv(a,d,&c,&r);
            zcopy(c,&a);
            zmod (a,d,&r);
         } while (ziszero(r)); 
         k++;
         if (!zscompare(a,1))
            break;
      }
   }
   if (zscompare(a,1))
   {  for (i=k;i--;)
         if (zcompare(a,pri[i]) > 0)  break;
      i++;
      for (j=k;j>i;j--)
      {  zcopy(pri[j-1],&pri[j]);
         exp[j]=exp[j-1];
      }
      zcopy(a,&pri[i]);
      exp[i] = 1;
      k++;
   }
   return  k ;
}


int  is_valid(long *n)
{
   return  !(n[0] <= 0  ||  n[0]==1 && n[1] <= 1);
}


int is_small_prime(unsigned long n)
{
   static unsigned char  *sieve=0;
   if (n <= 0)
      return 0;
   if (!sieve)
   {  unsigned  k, jj, hh, i;
      register unsigned  js, hs;
      if (!(sieve=calloc(PSIEVE_SIZE,1)))
         error_stop(8,"in is_small_prime: ");
      *sieve |= 1;   /* 1 is no prime */
      jj= 49/30;
      hh= 49%30;
      js= 0;
      hs= 2*7;
      for (i=2,k=2;jj<PSIEVE_SIZE;k+=2*i,i=3-i)
      {  if (k >= 5)
            k -= 5;
         if (k)
         {  register unsigned  j=jj, h=hh;
            while (j<PSIEVE_SIZE)
            {  switch (h)
               {  case  1:  sieve[j]|=1;   break;
                  case  7:  sieve[j]|=2;   break;
                  case 11:  sieve[j]|=4;   break;
                  case 13:  sieve[j]|=8;   break;
                  case 17:  sieve[j]|=16;  break;
                  case 19:  sieve[j]|=32;  break;
                  case 23:  sieve[j]|=64;  break;
                  case 29:  sieve[j]|=128; break;
               }
               j += js;
               h += hs;
               if (h >= 30)
                  h -= 30,  j++;
            }
         }
         jj += js<<i;
         hh += (hs+2*i)<<i;
         while (hh >= 30)
            hh -= 30,  jj++;
         hs += 4*i;
         if (hs >= 30)
            hs -= 30,  js++;
      }
   }
   if (n==2 || n==3 || n==5)
      return 1;
   if (n/30 >= PSIEVE_SIZE || !(n&1))
      return 0;
   switch (n%30)
   {  case  1:  return  !(sieve[n/30]&1);
      case  7:  return  !(sieve[n/30]&2);
      case 11:  return  !(sieve[n/30]&4);
      case 13:  return  !(sieve[n/30]&8);
      case 17:  return  !(sieve[n/30]&16);
      case 19:  return  !(sieve[n/30]&32);
      case 23:  return  !(sieve[n/30]&64);
      case 29:  return  !(sieve[n/30]&128);
      default:  return  0;
   }
}


int is_prime(verylong n, long probprime_tests)
{
   if (zscompare(n,30L*PSIEVE_SIZE) < 0)
      return  is_small_prime(ztoint(n));
   else
      return  !zmcomposite(n,probprime_tests);
}


void set_certificate(FILE *fp)
{
   certificat= fp;
}

#define  MAXFACTORS  64
int znm1ptest(verylong n)
{
   long  h, i, k, a, aa;
   unsigned  exp[MAXFACTORS];
   verylong  n1=0, pri[MAXFACTORS];
   static verylong  q=0, r=0, ma=0, mpow=0;
#ifdef  PTEST
   if ((h=zscompare(n,2L))<0)
      return 0;
   else if (!h)
      return 1;
   else if (!z2mod(n))
      return 0;
#else
   if (zscompare(n,30L*PSIEVE_SIZE)<0)
      return  is_small_prime(ztoint(n));
#endif
   for (i=0;i<MAXFACTORS;i++)
      pri[i]= 0;
   zsadd(n, -1L, &n1);
   k=factor(n1, pri, exp);
   if (k > MAXFACTORS)
      error_stop(6,"in znm1ptest: ");
   if (k <= 0)
      error_stop(13,"in znm1ptest: ");
   for (i=0;i<k;i++)
      if (!znm1ptest(pri[i]))
      {  fprintf(stderr,"!!  factor found composite prime: ");
         zwriteln(pri[i]);
         zwriteln(n1);
         error_stop(16,"in znm1ptest: ");
      }
   if (certificat)
   {  zfwrite(certificat,n);
      fprintf(certificat,":  N-1= ");
      for (i=0;i<k;i++)
      {  zfwrite(certificat,pri[i]);
         if (exp[i]>1)
            fprintf(certificat,"^%u",exp[i]); 
         fprintf(certificat," ");
      }
   }
   zmstart(n);
   aa=0;
   for (i=0;i<k;i++)
   {  zdiv(n1,pri[i],&q,&r);
      for (h=2,a= -1;;a+=h,h=6-h)
      {  zintom((a<=1?2+((a+1)>>1):a),&ma); 
         zmontexp(ma,q,&r);
         if (!zmisone(r))
         {  if (a==aa)
               break;
            aa=a;
            zmontexp(r,pri[i],&mpow);
            if (zmisone(mpow))
               break;
            zmfree();
            for (i=0;i<MAXFACTORS;i++)
               zfree(&pri[i]);
            zfree(&n1);
            return 0;
         }
      }
      if (certificat)
         fprintf(certificat," %u",(a<=1?2+((a+1)>>1):a));
   }
   if (certificat)
      fprintf(certificat,"\n");
   zmfree();
   for (i=0;i<MAXFACTORS;i++)
      zfree(&pri[i]);
   zfree(&n1);
   return 1;
}

static int u_lucas_is_zero(verylong q, long a, long b)
{
   unsigned  i;
#ifdef  NEW_LUCAS
   static verylong  v=0, v1=0, bQ=0, t0=0;
   zintom(a,&v);
   zintom(a*a-2*b,&v1);
   zintom(b,&bQ);
   for (i=z2log(q)-1; i-- ; )
   {  if (zbit(q,(long)i))
      {  zmontmul(v,v1,&v);
         zsmontmul(bQ,a,&t0);
         zmontsub(v,t0,&v);
         zmontsq(v1,&v1);
         zsmontmul(bQ,2*b,&t0);
         zmontsub(v1,t0,&v1);
         zmontsq(bQ,&bQ);
         zsmontmul(bQ,b,&bQ);
      }
      else
      {  zmontmul(v,v1,&v1);
         zsmontmul(bQ,a,&t0);
         zmontsub(v1,t0,&v1);
         zmontsq(v,&v);
         zmontsub(v,bQ,&v);
         zmontsub(v,bQ,&v);
         zmontsq(bQ,&bQ);
      }
   }
   zmontadd(v1,v1,&v1);
   zsmontmul(v,a,&v);
   zmontsub(v1,v,&v1);
   zintom(a*a-4*b,&t0);
   zmontinv(t0,&bQ);   /* could be stored */
   zmontmul(v1,bQ,&v);
   return  ziszero(v);
#else
   static verylong  u=0, v=0, bQ=0, t0=0, t1=0;
   long  d= a*a-4*b;
   zintom(1L,&u);
   zintom(a,&v);
   zintom(b,&bQ);
   for (i=z2log(q)-1; i-- ; )
   {  zmontmul(u,v,&u);
      zmontsq(v,&v);
      zmontsub(v,bQ,&v);
      zmontsub(v,bQ,&v);
      zmontsq(bQ,&bQ);
      if (zbit(q,(long)i))
      {  zcopy(v,&t0);
         zsmontmul(v,a,&v);
         zsmontmul(u,d,&t1);
         zmontadd(v,t1,&v);
         if (z2mod(v))
            zadd(v,zmmodulus(),&v);
         z2div(v,&v); 
         zsmontmul(u,a,&u);
         zmontadd(u,t0,&u);
         if (z2mod(u))
            zadd(u,zmmodulus(),&u);
         z2div(u,&u); 
         zsmontmul(bQ,b,&bQ);
      }
   }
   return  ziszero(u);
#endif
}

static long gcd(long a, long b)
{  /*  a >= 0 ,  b > 0 */
   long c;
   while (c=a%b)
      a=b, b=c;
   return  b;
}

int znp1ptest(verylong n)
{
   long  d, h, i, j, k;
   unsigned  exp[MAXFACTORS];
   verylong  n1=0, pri[MAXFACTORS];
#ifdef  PTEST
   if ((h=zscompare(n,2L))<0)
      return 0;
   else if (!h)
      return 1;
   else if (!z2mod(n))
      return 0;
#else
   if (zscompare(n,30L*PSIEVE_SIZE)<0)
      return  is_small_prime(ztoint(n));
#endif
   for (i=0;i<MAXFACTORS;i++)
      pri[i]= 0;
   zsadd(n, 1L, &n1);
   if ((i=lookup_primeaddr(n)) && !valid_factoring(i,1))
      get_factoring(i,1, &k, pri, exp);
   else
   {  k=factor(n1, pri, exp);
      if (k > MAXFACTORS)
         error_stop(6,"in znp1ptest: ");
      if (k <= 0)
         error_stop(13,"in znp1ptest: ");
   }
   for (i=0;i<k;i++)
      if (!znp1ptest(pri[i]))
      {  fprintf(stderr,"!!  factor found composite prime: ");
         zwriteln(pri[i]);
         zwriteln(n1);
         error_stop(16,"in znp1ptest: ");
      }
   if (certificat)
   {  zfwrite(certificat,n);
      fprintf(certificat,":  N+1= ");
      for (i=0;i<k;i++)
      {  zfwrite(certificat,pri[i]);
         if (exp[i]>1)
            fprintf(certificat,"^%u",exp[i]); 
         fprintf(certificat," ");
      }
   }
   h=4;
   zmstart(n);
   for (d=3;;d+=(d==3?2:h),h=6-h)
   {  if (d==25 || d==35)
         continue;
      if (!(j=zjacobis(zsmod(n,d),d)))
         break;
      if ((zslowbits(n,2L)&2&d)^(1-j))
      {  long a, b;
         static verylong  q=0, r=0;
         if (certificat)
            fprintf(certificat," D=%u ",(d&2?4*d:d));
         zdiv(n1,pri[i=0],&q,&r);
         a=(d&2?8:5);
         while (a*a <= (a&1?1:4)*d+4)
            a+=2+(d&2); 
         for (;;)
         {  b= (a*a-(a&1?1:4)*d)/4;
            for (j=(a+1)/2;j*j>b;j--); /* b be no square, else d=(a-2W)(a+2W) */
            if (j*j!=b && gcd(a,b)==1 && zscompare(n,b))
            {  if (gcd(zsmod(n,b),b) != 1)
                  goto out;
               j=0;
               while(!u_lucas_is_zero(q,a,b) && (j || u_lucas_is_zero(n1,a,b)))
               {  if (certificat)
                     fprintf(certificat," (%u,%u)",a,b);
                  j=1;
                  if (++i == k)
                  {  if (certificat)
                        fprintf(certificat,"\n");
                     zmfree();
                     for (i=0;i<MAXFACTORS;i++)
                        zfree(&pri[i]);
                     zfree(&n1);
                     return 1;
                  }
                  zdiv(n1,pri[i],&q,&r);
               }
            }
            a+=2+(d&2);
         }
      }
   }
   out:
   zmfree();
   for (i=0;i<MAXFACTORS;i++)
      zfree(&pri[i]);
   zfree(&n1);
   return 0;
}
#undef  MAXFACTORS


void  error_stop(int error_code, char *msg)
{
   fputs(msg,stderr);
   switch (error_code) 
   {  case  1:  fputs("invalid index\n",stderr);
      case  2:  fputs("table overflow\n",stderr);  break;
      case  3:  fputs("prime heap overflow\n",stderr);  break;
      case  4:  fputs("factorization list overflow\n",stderr);  break;
      case  5:  fputs("power heap overflow\n",stderr);  break;
      case  6:  fputs("too many factors. enlarge MAXFACTORS\n",stderr);  break;
      case  7:  fputs("no database defined\n",stderr);  break;
      case  8:  fputs("insufficient memory\n",stderr);  break;
      case  9:  fputs("no chain_factoring if ONE_UP defined\n",stderr); break;
      case 10:  fputs("invalid table_size\n",stderr); break;
      case 11:  fputs("invalid no of factorisations\n",stderr); break;
      case 12:  fputs("invalid prime_index in power_heap\n",stderr); break;
      case 13:  fputs("factor fails\n",stderr); break;
      default:  fprintf(stderr,"internal error %d\n",error_code); break;
   }
   exit(error_code);
}
